=== modified file 'NEWS'
--- NEWS	2009-01-28 21:57:12 +0000
+++ NEWS	2009-01-29 05:21:36 +0000
@@ -7,27 +7,6 @@
 IN DEVELOPMENT
 --------------
 
-  IMPROVEMENTS:
-
-    * Progress bars now show the rate of activity for some sftp 
-      operations, and they are drawn different.  (Martin Pool, #172741)
-
-  BUG FIXES:
-
-    * ``bzr annotate --show-ids`` doesn't give a backtrace on empty files
-      anymore.
-      (Anne Mohsen, Vincent Ladeuil, #314525)
-
-    * There was a bug in how we handled resolving when a file is deleted
-      in one branch, and modified in the other. If there was a criss-cross
-      merge, we would cause the deletion to conflict a second time.
-      (Vincent Ladeuil, John Arbash Meinel)
-
-    * There was another bug in how we chose the correct intermediate LCA in
-      criss-cross merges leading to several kind of changes be incorrectly
-      handled.
-      (John Arbash Meinel, Vincent Ladeuil)
-
   COMPATIBILITY BREAKS:
 
     * By default, ``bzr status`` after a merge now shows just the pending
@@ -42,6 +21,11 @@
 
   NEW FEATURES:
 
+    * Add support for filtering `bzr missing` on revisions.  Remote revisions
+      can be filtered using `bzr missing -r -20..-10` and local revisions can
+      be filtered using `bzr missing --my-revision -20..-10`.
+      (Marius Kruger)
+
     * ``bzr log -p`` displays the patch diff for each revision.
       When logging a file, the diff only includes changes to that file.
       (Ian Clatworthy, #202331, #227335)
@@ -57,6 +41,9 @@
       before the optional ``[merge]`` indicator. For line format,
       the tags appear after the date. (Ian Clatworthy)
 
+    * Progress bars now show the rate of activity for some sftp 
+      operations, and they are drawn different.  (Martin Pool, #172741)
+
     * Rule-based preferences can now accept multiple patterns for a set of
       rules.  (Marius Kruger)
 
@@ -66,6 +53,10 @@
 
   BUG FIXES:
 
+    * ``bzr annotate --show-ids`` doesn't give a backtrace on empty files
+      anymore.
+      (Anne Mohsen, Vincent Ladeuil, #314525)
+
     * ``bzr log FILE`` now correctly shows mainline revisions merging
       a change to FILE when the ``--short`` and ``--line`` log formats
       are used. (Ian Clatworthy, #317417)
@@ -82,6 +73,16 @@
       a new ``until_no_eintr`` helper function in ``bzrlib.osutils``.
       (Andrew Bennetts)
 
+    * There was a bug in how we handled resolving when a file is deleted
+      in one branch, and modified in the other. If there was a criss-cross
+      merge, we would cause the deletion to conflict a second time.
+      (Vincent Ladeuil, John Arbash Meinel)
+
+    * There was another bug in how we chose the correct intermediate LCA in
+      criss-cross merges leading to several kind of changes be incorrectly
+      handled.
+      (John Arbash Meinel, Vincent Ladeuil)
+
   DOCUMENTATION:
 
     * Improved plugin developer documentation.  (Martin Pool)

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2009-01-27 23:06:35 +0000
+++ bzrlib/builtins.py	2009-01-29 05:21:36 +0000
@@ -55,7 +55,13 @@
 """)
 
 from bzrlib.commands import Command, display_command
-from bzrlib.option import ListOption, Option, RegistryOption, custom_help
+from bzrlib.option import (
+    ListOption,
+    Option,
+    RegistryOption,
+    custom_help,
+    _parse_revision_str,
+    )
 from bzrlib.trace import mutter, note, warning, is_quiet, get_verbosity_level
 
 
@@ -1967,6 +1973,16 @@
             'bzr %s --revision takes one or two values.' % command_name)
     return rev1, rev2
 
+
+def _revision_range_to_revid_range(revision_range):
+    rev_id1 = None
+    rev_id2 = None
+    if revision_range[0] is not None:
+        rev_id1 = revision_range[0].rev_id
+    if revision_range[1] is not None:
+        rev_id2 = revision_range[1].rev_id
+    return rev_id1, rev_id2
+
 def get_log_format(long=False, short=False, line=False, default='long'):
     log_format = default
     if long:
@@ -3512,25 +3528,57 @@
 
 class cmd_missing(Command):
     """Show unmerged/unpulled revisions between two branches.
-    
+
     OTHER_BRANCH may be local or remote.
+
+    To filter on a range of revirions, you can use the command -r begin..end
+    -r revision requests a specific revision, -r ..end or -r begin.. are
+    also valid.
+
+    :Examples:
+
+        Determine the missing revisions between this and the branch at the
+        remembered pull location::
+
+            bzr missing
+
+        Determine the missing revisions between this and another branch::
+
+            bzr missing http://server/branch
+
+        Determine the missing revisions up to a specific revision on the other
+        branch::
+
+            bzr missing -r ..-10
+
+        Determine the missing revisions up to a specific revision on this
+        branch::
+
+            bzr missing --my-revision ..-10
     """
 
     _see_also = ['merge', 'pull']
     takes_args = ['other_branch?']
     takes_options = [
-            Option('reverse', 'Reverse the order of revisions.'),
-            Option('mine-only',
-                   'Display changes in the local branch only.'),
-            Option('this' , 'Same as --mine-only.'),
-            Option('theirs-only',
-                   'Display changes in the remote branch only.'),
-            Option('other', 'Same as --theirs-only.'),
-            'log-format',
-            'show-ids',
-            'verbose',
-            Option('include-merges', 'Show merged revisions.'),
-            ]
+        Option('reverse', 'Reverse the order of revisions.'),
+        Option('mine-only',
+               'Display changes in the local branch only.'),
+        Option('this' , 'Same as --mine-only.'),
+        Option('theirs-only',
+               'Display changes in the remote branch only.'),
+        Option('other', 'Same as --theirs-only.'),
+        'log-format',
+        'show-ids',
+        'verbose',
+        custom_help('revision',
+             help='Filter on other branch revisions (inclusive). '
+                'See "help revisionspec" for details.'),
+        Option('my-revision',
+            type=_parse_revision_str,
+            help='Filter on local branch revisions (inclusive). '
+                'See "help revisionspec" for details.'),
+        Option('include-merges', 'Show merged revisions.'),
+        ]
     encoding_type = 'replace'
 
     @display_command
@@ -3538,7 +3586,7 @@
             theirs_only=False,
             log_format=None, long=False, short=False, line=False,
             show_ids=False, verbose=False, this=False, other=False,
-            include_merges=False):
+            include_merges=False, revision=None, my_revision=None):
         from bzrlib.missing import find_unmerged, iter_log_revisions
         def message(s):
             if not is_quiet():
@@ -3572,6 +3620,15 @@
         remote_branch = Branch.open(other_branch)
         if remote_branch.base == local_branch.base:
             remote_branch = local_branch
+
+        local_revid_range = _revision_range_to_revid_range(
+            _get_revision_range(my_revision, local_branch,
+                self.name()))
+
+        remote_revid_range = _revision_range_to_revid_range(
+            _get_revision_range(revision,
+                remote_branch, self.name()))
+
         local_branch.lock_read()
         try:
             remote_branch.lock_read()
@@ -3579,7 +3636,9 @@
                 local_extra, remote_extra = find_unmerged(
                     local_branch, remote_branch, restrict,
                     backward=not reverse,
-                    include_merges=include_merges)
+                    include_merges=include_merges,
+                    local_revid_range=local_revid_range,
+                    remote_revid_range=remote_revid_range)
 
                 if log_format is None:
                     registry = log.log_formatter_registry
@@ -4716,10 +4775,7 @@
                 revid1, revid2 = rev1.rev_id, rev2.rev_id
                 # only show revisions between revid1 and revid2 (inclusive)
                 tags = [(tag, revid) for tag, revid in tags if
-                     (revid2 is None or
-                         graph.is_ancestor(revid, revid2)) and
-                     (revid1 is None or
-                         graph.is_ancestor(revid1, revid))]
+                    graph.is_between(revid, revid1, revid2)]
             finally:
                 branch.unlock()
         if sort == 'alpha':

=== modified file 'bzrlib/graph.py'
--- bzrlib/graph.py	2008-12-12 01:17:04 +0000
+++ bzrlib/graph.py	2009-01-01 21:46:06 +0000
@@ -920,6 +920,17 @@
         return set([candidate_descendant]) == self.heads(
             [candidate_ancestor, candidate_descendant])
 
+    def is_between(self, revid, lower_bound_revid, upper_bound_revid):
+        """Determine whether a revision is between two others.
+
+        returns true if and only if:
+        lower_bound_revid <= revid <= upper_bound_revid
+        """
+        return ((upper_bound_revid is None or
+                    self.is_ancestor(revid, upper_bound_revid)) and
+               (lower_bound_revid is None or
+                    self.is_ancestor(lower_bound_revid, revid)))
+
     def _search_for_extra_common(self, common, searchers):
         """Make sure that unique nodes are genuinely unique.
 

=== modified file 'bzrlib/missing.py'
--- bzrlib/missing.py	2008-09-11 19:36:38 +0000
+++ bzrlib/missing.py	2009-01-01 23:21:07 +0000
@@ -53,7 +53,8 @@
 
 
 def find_unmerged(local_branch, remote_branch, restrict='all',
-                  include_merges=False, backward=False):
+                  include_merges=False, backward=False,
+                  local_revid_range=None, remote_revid_range=None):
     """Find revisions from each side that have not been merged.
 
     :param local_branch: Compare the history of local_branch
@@ -66,7 +67,11 @@
     :param include_merges: Show mainline revisions only if False,
         all revisions otherwise.
     :param backward: Show oldest versions first when True, newest versions
-        first when False. 
+        first when False.
+    :param local_revid_range: Revision-id range for filtering local_branch
+        revisions (lower bound, upper bound)
+    :param remote_revid_range: Revision-id range for filtering remote_branch
+        revisions (lower bound, upper bound)
 
     :return: A list of [(revno, revision_id)] for the mainline revisions on
         each side.
@@ -77,7 +82,9 @@
         try:
             return _find_unmerged(
                 local_branch, remote_branch, restrict=restrict,
-                include_merges=include_merges, backward=backward)
+                include_merges=include_merges, backward=backward,
+                local_revid_range=local_revid_range,
+                remote_revid_range=remote_revid_range)
         finally:
             remote_branch.unlock()
     finally:
@@ -174,8 +181,16 @@
     return revline
 
 
+def _filter_revs(graph, revs, revid_range):
+    if revid_range is None or revs is None:
+        return revs
+    return [rev for rev in revs
+        if graph.is_between(rev[1], revid_range[0], revid_range[1])]
+
+
 def _find_unmerged(local_branch, remote_branch, restrict,
-                   include_merges, backward):
+                   include_merges, backward,
+                   local_revid_range=None, remote_revid_range=None):
     """See find_unmerged.
 
     The branches should already be locked before entering.
@@ -214,7 +229,8 @@
                                      local_revision_id, backward)
         remotes = _enumerate_mainline(remote_extra, graph, remote_revno,
                                       remote_revision_id, backward)
-    return locals, remotes
+    return _filter_revs(graph, locals, local_revid_range), _filter_revs(graph,
+        remotes, remote_revid_range)
 
 
 def sorted_revisions(revisions, history_map):

=== modified file 'bzrlib/tests/blackbox/test_missing.py'
--- bzrlib/tests/blackbox/test_missing.py	2008-10-17 08:32:52 +0000
+++ bzrlib/tests/blackbox/test_missing.py	2009-01-25 00:55:26 +0000
@@ -26,6 +26,13 @@
 
 class TestMissing(TestCaseWithTransport):
 
+    def assertMessages(self, out, must_have=(), must_not_have=()):
+        """Check if commit messages are in or not in the output"""
+        for m in must_have:
+            self.assertContainsRe(out, r'\nmessage:\n  %s\n' % m)
+        for m in must_not_have:
+            self.assertNotContainsRe(out, r'\nmessage:\n  %s\n' % m)
+
     def test_missing_quiet(self):
         # <https://bugs.launchpad.net/bzr/+bug/284748>
         # create a source branch
@@ -137,6 +144,40 @@
         self.assertEqualDiff('Other branch is up to date.\n',
                              self.run_bzr('missing ../a --theirs-only')[0])
 
+    def test_missing_filtered(self):
+        # create a source branch
+        a_tree = self.make_branch_and_tree('a')
+        self.build_tree_contents([('a/a', 'initial\n')])
+        a_tree.add('a')
+        a_tree.commit(message='r1')
+        # clone and add differing revisions
+        b_tree = a_tree.bzrdir.sprout('b').open_workingtree()
+
+        for i in range(2, 6):
+            a_tree.commit(message='a%d' % i)
+            b_tree.commit(message='b%d' % i)
+
+        os.chdir('a')
+        # local
+        out,err = self.run_bzr('missing ../b --my-revision 3', retcode=1)
+        self.assertMessages(out, ('a3', 'b2', 'b3', 'b4', 'b5'), ('a2', 'a4'))
+
+        out,err = self.run_bzr('missing ../b --my-revision 3..4', retcode=1)
+        self.assertMessages(out, ('a3', 'a4'), ('a2', 'a5'))
+
+        #remote
+        out,err = self.run_bzr('missing ../b -r 3', retcode=1)
+        self.assertMessages(out, ('a2', 'a3', 'a4', 'a5', 'b3'), ('b2', 'b4'))
+
+        out,err = self.run_bzr('missing ../b -r 3..4', retcode=1)
+        self.assertMessages(out, ('b3', 'b4'), ('b2', 'b5'))
+
+        #both
+        out,err = self.run_bzr('missing ../b --my-revision 3..4 -r 3..4',
+            retcode=1)
+        self.assertMessages(out, ('a3', 'a4', 'b3', 'b4'),
+            ('a2', 'a5', 'b2', 'b5'))
+
     def test_missing_check_last_location(self):
         # check that last location shown as filepath not file URL
 

=== modified file 'bzrlib/tests/test_graph.py'
--- bzrlib/tests/test_graph.py	2008-12-11 16:25:46 +0000
+++ bzrlib/tests/test_graph.py	2009-01-01 21:46:06 +0000
@@ -698,6 +698,17 @@
         instrumented_graph.is_ancestor('rev2a', 'rev2b')
         self.assertTrue('null:' not in instrumented_provider.calls)
 
+    def test_is_between(self):
+        graph = self.make_graph(ancestry_1)
+        self.assertEqual(True, graph.is_between('null:', 'null:', 'null:'))
+        self.assertEqual(True, graph.is_between('rev1', 'null:', 'rev1'))
+        self.assertEqual(True, graph.is_between('rev1', 'rev1', 'rev4'))
+        self.assertEqual(True, graph.is_between('rev4', 'rev1', 'rev4'))
+        self.assertEqual(True, graph.is_between('rev3', 'rev1', 'rev4'))
+        self.assertEqual(False, graph.is_between('rev4', 'rev1', 'rev3'))
+        self.assertEqual(False, graph.is_between('rev1', 'rev2a', 'rev4'))
+        self.assertEqual(False, graph.is_between('null:', 'rev1', 'rev4'))
+
     def test_is_ancestor_boundary(self):
         """Ensure that we avoid searching the whole graph.
         

=== modified file 'bzrlib/tests/test_missing.py'
--- bzrlib/tests/test_missing.py	2008-09-12 08:44:07 +0000
+++ bzrlib/tests/test_missing.py	2009-01-25 01:03:47 +0000
@@ -138,13 +138,14 @@
 class TestFindUnmerged(tests.TestCaseWithTransport):
 
     def assertUnmerged(self, local, remote, local_branch, remote_branch,
-                       restrict, include_merges=False,
-                       backward=False):
+            restrict='all', include_merges=False, backward=False,
+            local_revid_range=None, remote_revid_range=None):
         """Check the output of find_unmerged_mainline_revisions"""
         local_extra, remote_extra = missing.find_unmerged(
-                                        local_branch, remote_branch, restrict,
-                                        include_merges=include_merges,
-                                        backward=backward)
+            local_branch, remote_branch, restrict,
+            include_merges=include_merges, backward=backward,
+            local_revid_range=local_revid_range,
+            remote_revid_range=remote_revid_range)
         self.assertEqual(local, local_extra)
         self.assertEqual(remote, remote_extra)
 
@@ -153,22 +154,24 @@
         rev1 = tree.commit('one')
         tree.lock_read()
         self.addCleanup(tree.unlock)
-        self.assertUnmerged([], [], tree.branch, tree.branch, 'all')
+        self.assertUnmerged([], [], tree.branch, tree.branch)
+        self.assertUnmerged([], [], tree.branch, tree.branch,
+            local_revid_range=(rev1, rev1))
 
     def test_one_ahead(self):
         tree = self.make_branch_and_tree('tree')
         rev1 = tree.commit('one')
         tree2 = tree.bzrdir.sprout('tree2').open_workingtree()
         rev2 = tree2.commit('two')
-        self.assertUnmerged([], [('2', rev2)], tree.branch, tree2.branch, 'all')
-        self.assertUnmerged([('2', rev2)], [], tree2.branch, tree.branch, 'all')
+        self.assertUnmerged([], [('2', rev2)], tree.branch, tree2.branch)
+        self.assertUnmerged([('2', rev2)], [], tree2.branch, tree.branch)
 
     def test_restrict(self):
         tree = self.make_branch_and_tree('tree')
         rev1 = tree.commit('one')
         tree2 = tree.bzrdir.sprout('tree2').open_workingtree()
         rev2 = tree2.commit('two')
-        self.assertUnmerged([], [('2', rev2)], tree.branch, tree2.branch, 'all')
+        self.assertUnmerged([], [('2', rev2)], tree.branch, tree2.branch)
         self.assertUnmerged([], None, tree.branch, tree2.branch, 'local')
         self.assertUnmerged(None, [('2', rev2)], tree.branch, tree2.branch,
                                                'remote')
@@ -182,7 +185,11 @@
         tree.merge_from_branch(tree2.branch)
         rev4 = tree.commit('four')
 
-        self.assertUnmerged([('2', rev4)], [], tree.branch, tree2.branch, 'all')
+        self.assertUnmerged([('2', rev4)], [], tree.branch, tree2.branch)
+        self.assertUnmerged([('2', rev4)], [], tree.branch, tree2.branch,
+            local_revid_range=(rev4, rev4))
+        self.assertUnmerged([], [], tree.branch, tree2.branch,
+            local_revid_range=(rev1, rev1))
 
     def test_include_merges(self):
         tree = self.make_branch_and_tree('tree')
@@ -199,17 +206,80 @@
         tree2.merge_from_branch(tree3.branch)
         rev6 = tree2.commit('six', rev_id='rev6')
 
-        self.assertUnmerged([], [('2', 'rev2', 0), ('3', 'rev3',0 ),
+        self.assertUnmerged([], [('2', 'rev2', 0), ('3', 'rev3', 0),
                                  ('4', 'rev6', 0),
                                  ('3.1.1', 'rev4', 1), ('3.1.2', 'rev5', 1),
                                  ],
-                            tree.branch, tree2.branch, 'all',
+                            tree.branch, tree2.branch,
                             include_merges=True)
 
         self.assertUnmerged([], [('4', 'rev6', 0),
                                  ('3.1.2', 'rev5', 1), ('3.1.1', 'rev4', 1),
-                                 ('3', 'rev3',0 ), ('2', 'rev2', 0),
+                                 ('3', 'rev3', 0), ('2', 'rev2', 0),
                                  ],
-                            tree.branch, tree2.branch, 'all',
+                            tree.branch, tree2.branch,
                             include_merges=True,
                             backward=True)
+
+        self.assertUnmerged([], [('4', 'rev6', 0)],
+            tree.branch, tree2.branch,
+            include_merges=True, remote_revid_range=(rev6, rev6))
+
+        self.assertUnmerged([], [('3', 'rev3', 0), ('3.1.1', 'rev4', 1)],
+                    tree.branch, tree2.branch,
+                    include_merges=True, remote_revid_range=(rev3, rev4))
+
+        self.assertUnmerged([], [('4', 'rev6', 0), ('3.1.2', 'rev5', 1)],
+                    tree.branch, tree2.branch,
+                    include_merges=True, remote_revid_range=(rev5, rev6))
+
+    def test_revision_range(self):
+        local = self.make_branch_and_tree('local')
+        lrevid1 = local.commit('one')
+        remote = local.bzrdir.sprout('remote').open_workingtree()
+        rrevid2 = remote.commit('two')
+        rrevid3 = remote.commit('three')
+        rrevid4 = remote.commit('four')
+        lrevid2 = local.commit('two')
+        lrevid3 = local.commit('three')
+        lrevid4 = local.commit('four')
+        local_extra = [('2', lrevid2), ('3', lrevid3), ('4', lrevid4)]
+        remote_extra = [('2', rrevid2), ('3', rrevid3), ('4', rrevid4)]
+
+        # control
+        self.assertUnmerged(local_extra, remote_extra,
+            local.branch, remote.branch)
+        self.assertUnmerged(local_extra, remote_extra,
+            local.branch, remote.branch, local_revid_range=(None, None),
+            remote_revid_range=(None, None))
+
+        # exclude local revisions
+        self.assertUnmerged([('2', lrevid2)], remote_extra,
+            local.branch, remote.branch, local_revid_range=(lrevid2, lrevid2))
+        self.assertUnmerged([('2', lrevid2), ('3', lrevid3)], remote_extra,
+            local.branch, remote.branch, local_revid_range=(lrevid2, lrevid3))
+        self.assertUnmerged([('2', lrevid2), ('3', lrevid3)], None,
+            local.branch, remote.branch, 'local',
+            local_revid_range=(lrevid2, lrevid3))
+
+        # exclude remote revisions
+        self.assertUnmerged(local_extra, [('2', rrevid2)],
+            local.branch, remote.branch, remote_revid_range=(None, rrevid2))
+        self.assertUnmerged(local_extra, [('2', rrevid2)],
+            local.branch, remote.branch, remote_revid_range=(lrevid1, rrevid2))
+        self.assertUnmerged(local_extra, [('2', rrevid2)],
+            local.branch, remote.branch, remote_revid_range=(rrevid2, rrevid2))
+        self.assertUnmerged(local_extra, [('2', rrevid2), ('3', rrevid3)],
+            local.branch, remote.branch, remote_revid_range=(None, rrevid3))
+        self.assertUnmerged(local_extra, [('2', rrevid2), ('3', rrevid3)],
+            local.branch, remote.branch, remote_revid_range=(rrevid2, rrevid3))
+        self.assertUnmerged(local_extra, [('3', rrevid3)],
+            local.branch, remote.branch, remote_revid_range=(rrevid3, rrevid3))
+        self.assertUnmerged(None, [('2', rrevid2), ('3', rrevid3)],
+            local.branch, remote.branch, 'remote',
+            remote_revid_range=(rrevid2, rrevid3))
+
+        # exclude local and remote revisions
+        self.assertUnmerged([('3', lrevid3)], [('3', rrevid3)],
+            local.branch, remote.branch, local_revid_range=(lrevid3, lrevid3),
+            remote_revid_range=(rrevid3, rrevid3))

